// -------------------------------------------------------------
// Terrain Shader
// + prt
// + dynamic texturing
// + reflection rendering
// + shadow mapping
// -------------------------------------------------------------

// -------------------------------------------------------------
// common values

float4x4	matWorldViewProj;
float4x4	matWorld;
float4		cLight;

struct VS_OUTPUT
{
	float4	Position	: POSITION;
	float4	Diffuse		: COLOR0;
	float2	TexCoord	: TEXCOORD0;
	float3	ClipDist	: TEXCOORD1;
	float	Height		: TEXCOORD2;
	float	Angle		: TEXCOORD3;
};

struct PS_OUTPUT
{
	float4	RGBAColor	: COLOR0;	
};
// -------------------------------------------------------------


// -------------------------------------------------------------
// reflection

float4		clipPlane;
// -------------------------------------------------------------



// -------------------------------------------------------------
// PRT

// NumCoefficients is defined by app upon Effect.FromFile
// #define NumCoefficients	36

float		lightScale;
float4		vLightCoefficients[NumCoefficients/4];

// Calculates and clamps the brightness of a vertex using PRT
float GetBrightness(float4 vSHCoefficients[NumCoefficients/4])
{
	float4	vBrightness = float4(0,0,0,0);
	
	for(int j=0; j < (NumCoefficients/4); j++)
	{
		vBrightness += vLightCoefficients[j] * vSHCoefficients[j];
	}
		
	float	fBrightness = dot(vBrightness,1);
	fBrightness = saturate(fBrightness);
	
	return fBrightness;
}
// -------------------------------------------------------------


// -------------------------------------------------------------
// dynamic texturing

const float3	upVec = float3(0.0f, 1.0f, 0.0f);

float		heightScale;

float		trans1Factor;
float		trans1End;

float		angleFactor;
float		steep1Factor;
float		steep2Factor;
float		steep1Begin;
float		steep2End;
float		angleBegin;

texture		heightTexture1;
texture		heightTexture2;
texture		steepTexture;

sampler sampTexture1 = sampler_state
{ 
    Texture		= (heightTexture1);
    MipFilter	= LINEAR; 
    MinFilter	= LINEAR;
    MagFilter	= LINEAR;
    AddressU	= WRAP;
    AddressV	= WRAP;
};

sampler sampTexture2 = sampler_state
{ 
    Texture		= (heightTexture2);
    MipFilter	= LINEAR; 
    MinFilter	= LINEAR;
    MagFilter	= LINEAR;
    AddressU	= WRAP;
    AddressV	= WRAP;
};

sampler sampSteepTexture = sampler_state
{ 
    Texture		= (steepTexture);
    MipFilter	= LINEAR; 
    MinFilter	= LINEAR;
    MagFilter	= LINEAR;
    AddressU	= WRAP;
    AddressV	= WRAP;
};
// -------------------------------------------------------------


// -------------------------------------------------------------
// shadow mapping

// SHADOWMAPSIZE is defined by app upon Effect.FromFile
// #define	SHADOWMAPSIZE	2048.0f

// SHADOWEPSILON is defined by app upon Effect.FromFile
// #define	SHADOWEPSILON	0.00005f

float4x4	matWorldView;
float4x4	matProj;
float4x4	matViewToLightProj;
float4x4	matTextureGen;
float4x4	matLightViewProj;

float		shadowIntensity;
float4		shadowColor;
float4		lightDirection;

texture		shadowMapTexture;

sampler shadowMapSampler = sampler_state
{
	Texture		= (shadowMapTexture);
	MinFilter	= Point;
	MagFilter	= Point;
	MipFilter	= NONE;
	AddressU	= Clamp;
	AddressV	= Clamp;
};

struct VS_SHADOWOUTPUT
{
	float4	Position	: POSITION;
	float4	Diffuse		: COLOR0;
	float2	TexCoord	: TEXCOORD0;
	float	Height		: TEXCOORD1;
	float	Angle		: TEXCOORD2;
	float4	PosLight	: TEXCOORD3;
};
// -------------------------------------------------------------


//-----------------------------------------------------------------------------
// Vertex Shader
// + per vertex PRT
// + calculates the steepness
// + calculates clipdistance
//-----------------------------------------------------------------------------
VS_OUTPUT PrtDiffuseVS	(	float4			vPos								: POSITION,
							float3			vNormal								: NORMAL,
							float2			TexCoord							: TEXCOORD0,
							float4			vSHCoefficients[NumCoefficients/4]	: BLENDWEIGHT,
							uniform bool	bReflect
						)
{
	VS_OUTPUT	Output;
	
	Output.Position	= mul(vPos, matWorldViewProj);
	Output.Height	= vPos.y/heightScale;	
	Output.TexCoord	= TexCoord;

		
	if(bReflect)
	{
		Output.ClipDist = dot(vPos, clipPlane.xyz) - clipPlane.w;
	}
	else
	{
		Output.ClipDist = 0;
	}
		
		
	float	fBrightness	= saturate(GetBrightness(vSHCoefficients) * lightScale);
	Output.Diffuse		= cLight * fBrightness;
	
	// Winkel ausrechnen (1.0f = 90, 0.0f = 0)
	Output.Angle	= 1.0f - abs(dot(vNormal, upVec));
	
	return Output;
}

//-----------------------------------------------------------------------------
// Pixel Shader
// + clips
// + calculates transparency of the terrain
// + textures the terrain depending on steepness and height
//-----------------------------------------------------------------------------
PS_OUTPUT PrtDiffusePS	(	VS_OUTPUT		In,
							uniform bool	bReflect
						)
{
	PS_OUTPUT	Output;

	float waterDepth;
	
	if( bReflect )
	{
		clip(In.ClipDist);
		waterDepth = 0;
	}
	else
	{	
		waterDepth = (clipPlane.w - In.Height*heightScale)/clipPlane.w;
	}
	
		
	float	tex1Factor	= saturate((trans1End - In.Height)*trans1Factor);		
	float4	groundColor	= lerp(tex2D(sampTexture2, In.TexCoord), tex2D(sampTexture1, In.TexCoord), tex1Factor);
	
	//float	steepFactor_up	= (In.Height - steep1Begin)*steep1Factor;
	//float	steepFactor_dn	= (steep2End - In.Height)*steep2Factor;
	//float	steepFactor	= saturate(min(steepFactor_up, steepFactor_dn));
	
	float	steepAngleFactor = saturate((In.Angle - angleBegin)*angleFactor);
	
	//float	a = steepFactor * steepAngleFactor;
	//float4 steepColor = a * tex2D(sampSteepTexture, In.TexCoord);	
	float4	steepColor = steepAngleFactor * tex2D(sampSteepTexture, In.TexCoord);	
	
	//Output.RGBAColor	= In.Diffuse * lerp(groundColor, steepColor, a);
	Output.RGBAColor	= In.Diffuse * lerp(groundColor, steepColor, steepAngleFactor);
	Output.RGBAColor.w	= 1.0f - waterDepth;
	
    return Output;
}

//-----------------------------------------------------------------------------
// Vertex Shader
// + per vertex PRT
// + calculates the steepness
// + calculates clipdistance
// + prepares for shadow mapping
//-----------------------------------------------------------------------------
VS_SHADOWOUTPUT PrtDiffuseShadowMapVS	(	float4			vPos								: POSITION,
											float3			vNormal								: NORMAL,
											float2			TexCoord							: TEXCOORD0,
											float4			vSHCoefficients[NumCoefficients/4]	: BLENDWEIGHT
										)
{
	VS_SHADOWOUTPUT	Output;
	
	float4	tmpPos		= mul(vPos, matWorldView);
	float	fBrightness = saturate(GetBrightness(vSHCoefficients) * lightScale);

	Output.Position		= mul(tmpPos, matProj);	
	Output.Height		= vPos.y/heightScale;	
	Output.TexCoord		= TexCoord;
	Output.Diffuse		= cLight * fBrightness;
	
	// Winkel ausrechnen (1.0f = 90, 0.0f = 0)
	Output.Angle	= 1.0f - abs(dot(vNormal, upVec));
	
	// Generate coordinates for shadow mapping
	Output.PosLight		= mul(tmpPos, matViewToLightProj);
	
	return Output;
}

//-----------------------------------------------------------------------------
// Pixel Shader
// + clips
// + calculates transparency of the terrain
// + textures the terrain depending on steepness and height
// + shadow mapping
//-----------------------------------------------------------------------------
PS_OUTPUT PrtDiffuseShadowMapPS	(VS_SHADOWOUTPUT In)
{
	PS_OUTPUT	Output;
	
	
	float2	shadowTexCoord	= 0.5f * In.PosLight.xy / In.PosLight.w + float2(0.5, 0.5);
	shadowTexCoord.y = 1.0f - shadowTexCoord.y;
	
	// transform to texel space
    float2 texelpos = SHADOWMAPSIZE * shadowTexCoord;
    
    // Determine the lerp amounts           
    float2 lerps = frac( texelpos );

	float sourcevals[4];
    sourcevals[0] = (tex2D( shadowMapSampler, shadowTexCoord )                                                 + SHADOWEPSILON < In.PosLight.z / In.PosLight.w)? 0.0f: 0.8f;//1.0f;
    sourcevals[1] = (tex2D( shadowMapSampler, shadowTexCoord + float2(1.0f/SHADOWMAPSIZE, 0                ) ) + SHADOWEPSILON < In.PosLight.z / In.PosLight.w)? 0.0f: 0.8f;//1.0f;  
    sourcevals[2] = (tex2D( shadowMapSampler, shadowTexCoord + float2(0.0f,               1.0/SHADOWMAPSIZE) ) + SHADOWEPSILON < In.PosLight.z / In.PosLight.w)? 0.0f: 0.8f;//1.0f;  
    sourcevals[3] = (tex2D( shadowMapSampler, shadowTexCoord + float2(1.0f/SHADOWMAPSIZE, 1.0/SHADOWMAPSIZE) ) + SHADOWEPSILON < In.PosLight.z / In.PosLight.w)? 0.0f: 0.8f;//1.0f;  		

	// lerp between the shadow values to calculate our light amount
    float lightAmount = lerp(	lerp( sourcevals[0], sourcevals[1], lerps.x ),
                                lerp( sourcevals[2], sourcevals[3], lerps.x ),
                                lerps.y );
                                
	float4	baseColor = lerp(In.Diffuse, shadowColor, (1.0f-lightAmount)*shadowIntensity*0.7f);
  
	float	waterDepth = (clipPlane.w - In.Height*heightScale)/clipPlane.w;

	float	tex1Factor		= saturate((trans1End - In.Height)*trans1Factor);		
	float4	groundColor	= lerp(tex2D(sampTexture2, In.TexCoord), tex2D(sampTexture1, In.TexCoord), tex1Factor);
	
	float	steepFactor_up	= (In.Height - steep1Begin)*steep1Factor;
	float	steepFactor_dn	= (steep2End - In.Height)*steep2Factor;
	float	steepFactor		= saturate(min(steepFactor_up, steepFactor_dn));
	
	float	steepAngleFactor = saturate((In.Angle - angleBegin)*angleFactor);
	
	float	a			= steepFactor * steepAngleFactor;
	float4	steepColor	= a * tex2D(sampSteepTexture, In.TexCoord);			

	Output.RGBAColor	= baseColor * lerp(groundColor, steepColor, a);
	Output.RGBAColor.w	= 1.0f - waterDepth;	
	
    return Output;
}


// -------------------------------------------------------------
// Name technique | pass | how to compile 
// -------------------------------------------------------------
technique RenderWithPrt
{
    pass P0
	{
        VertexShader = compile vs_1_1 PrtDiffuseVS(false);
        PixelShader  = compile ps_2_0 PrtDiffusePS(false);
    }
}

technique RenderReflectionWithPrt
{
    pass P0
	{
        VertexShader = compile vs_1_1 PrtDiffuseVS(true);
        PixelShader  = compile ps_2_0 PrtDiffusePS(true);
    }
}

technique RenderWithPrtShadowMapped
{
	pass P0
	{
		VertexShader = compile vs_1_1 PrtDiffuseShadowMapVS();
		PixelShader  = compile ps_2_0 PrtDiffuseShadowMapPS();
	}
}